"""
Axon population generator usefull functions.
"""
import faulthandler 
import math
import os
from rich.progress import track

import matplotlib.pyplot as plt
from itertools import combinations
import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import rv_continuous, gamma

from ..backend._log_interface import pass_info, progression_popup, rise_warning
from ..backend._parameters import parameters

# WARNING:
# no prompt message for numpy division by zeros: handled in the code !!!
np.seterr(divide="ignore", invalid="ignore")

# verbosity level
fg_verbose = True

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()

# get the built-in material librairy
dir_path = parameters.nrv_path + "/_misc"
stat_library = os.listdir(dir_path + "/stats/")

#################################################
## Nerve composition statistics and generators ##
#################################################
myelinated_stats = [
    "Schellens_1",
    "Schellens_2",
    "Ochoa_M",
    "Jacobs_9_A",
    "Jacobs_9_B",
    "Jacobs_9_C",
    "Jacobs_9_D",
]
unmyelinated_stats = [
    "Ochoa_U",
    "Jacobs_11_A",
    "Jacobs_11_B",
    "Jacobs_11_C",
    "Jacobs_11_D",
]


def load_stat(stat_name):
    """
    Load a statistic stored in the included librairy or specified by the user in a csv file (first corlumn: axon diameter bin start - second corlumn: proportion)

    Parameters
    ----------
    stat_name : str
        name of the statistic in the librairy or path to a new librairy in csv

    Returns
    -------
    diameters   : list
        diameters as start of bins of the histogram
    presence    : list
        quantities of presence of each bin in the histogram
    """
    f_in_librairy = str(stat_name) + ".csv"
    if f_in_librairy in stat_library:
        stat_file = np.genfromtxt(dir_path + "/stats/" + f_in_librairy, delimiter=",")
    else:
        # stat_file = np.genfromtxt(dir_path + "/stats/" + f_in_librairy, delimiter=",")
        stat_file = np.genfromtxt(f_in_librairy, delimiter=",")
    diameters = stat_file[:, 0]
    presence = stat_file[:, 1]
    return diameters, presence


def one_Gamma(x, a1, beta1, c1):
    """
    Gamma function, mono-modal, to interpolate unmyelinated statistics.

    Parameters
    ----------
    x       : float
        variable of the Gamma function, will correspond to diameter
    a1      : float
        shape parameter
    beta1   : float
        scale parameter
    c1      : float
        gain applied to the gamma function

    Note
    ----
    location of the gamma pdf function is fixed to 0.2, diameters will be interpolated above this minimal value.
    """
    return c1 * (gamma.pdf(x, a1, scale=1 / beta1, loc=0.2))


def two_Gamma(x, a1, beta1, c1, a2, beta2, c2, transition):
    """
    Gamma function, bi-modal, to interpolate myelinated statistics.

    Parameters
    ----------
    x           : float
        variable of the Gamma function, will correspond to diameter
    a1          : float
        shape parameter on the first (lowest) lobe
    beta1       : float
        scale parameter on the first (lowest) lobe
    c1          : float
        gain applied to the gamma function on the first (lowest) lobe
    a2          : float
        shape parameter on the second (highest) lobe
    beta2       : float
        scale parameter on the second (highest) lobe
    c2          : float
        gain applied to the gamma function on the second (highest) lobe
    transition  :float
        diameter value at the transition between the two lobes

    Note
    ----
    location of the gamma pdf first function is fixed to 2, diameters will be interpolated above this minimal value. This corresponds to the minimal A-delta diamter tolerated value.
    """
    return c1 * (gamma.pdf(x, a1, scale=1 / beta1, loc=2)) + c2 * (
        gamma.pdf(x, a2, scale=1 / beta2, loc=transition)
    )


class nerve_gen_one_gamma(rv_continuous):
    """
    Class to create specific repartition law generator for unmyelinated axons. Inherits from scipy rv_continuous class. Some methods are overwritten
    """

    def set_bounds(self, v_min, v_max):
        """
        Set bounds to the generator.

        Parameters
        ----------
        v_min   : float
            minimum value
        v_max   : float
            maximum value
        """
        self.a = v_min
        self.b = v_max

    def configure(self, a1, beta1, c1):
        """
        Set the value from interpolated data

        Parameters
        ----------
        a1      : float
            shape parameter
        beta1   : float
            scale parameter
        c1      : float
            gain applied to the gamma function

        """
        self.a1 = a1
        self.beta1 = beta1
        self.c1 = c1

    def _pdf(self, x):
        """
        overwritting the pdf to custom law, same definition as one_Gamma described upper.
        """
        return self.c1 * (gamma.pdf(x, self.a1, scale=1 / self.beta1, loc=0.2))


class nerve_gen_two_gamma(rv_continuous):
    """
    Class to create specific repartition law generator for myelinated axons. Inherits from scipy rv_continuous class. Some methods are overwritten
    """

    def set_bounds(self, v_min, v_max):
        """
        Set bounds to the generator.

        Parameters
        ----------
        v_min   : float
            minimum value
        v_max   : float
            maximum value
        """
        self.a = v_min
        self.b = v_max

    def configure(self, a1, beta1, c1, a2, beta2, c2, transition):
        """
        Set the value from interpolated data

        Parameters
        ----------
        a1          : float
            shape parameter on the first (lowest) lobe
        beta1       : float
            scale parameter on the first (lowest) lobe
        c1          : float
            gain applied to the gamma function on the first (lowest) lobe
        a2          : float
            shape parameter on the second (highest) lobe
        beta2       : float
            scale parameter on the second (highest) lobe
        c2          : float
            gain applied to the gamma function on the second (highest) lobe
        transition  :float
            diameter value at the transition between the two lobes

        """
        self.a1 = a1
        self.a2 = a2
        self.beta1 = beta1
        self.beta2 = beta2
        self.c1 = c1
        self.c2 = c2
        self.transition = transition

    def _pdf(self, x):
        """
        overwritting the pdf to custom law, same definition as two_Gamma described upper.
        """
        return self.c1 * (
            gamma.pdf(x, self.a1, scale=1 / self.beta1, loc=2)
        ) + self.c2 * (
            gamma.pdf(x, self.a2, scale=1 / self.beta2, loc=self.transition)
        )


def create_generator_from_stat(stat, myelinated=True, dmin=None, dmax=None, one_gamma:bool = False):
    """
    Create a statistical generator (type rv_continuous) for a given statistic.

    Parameters
    ----------
    stat        : str
        name of the statistic in the librairy or path to a new librairy in csv
    myelinated  : bool
        True if the statistic is describing myelinated diameters, esle False. If the statisticis in the librairy, this will be automatically chosen.
    dmin        : float
        minimal diameter to consider, in um. If None, the minimal value of the statistic is taken
    dmax        : float
        minimal diameter to consider, in um. If None, the maximal value of the statistic plus bin size is taken
    one_gamma    : bool
        If true, the stat generator is forced to one gamma only
        
    Returns
    -------
    generator   : rv_continuous
        Generator object corresponding to the statistic.
    """
    # Check if myelinated or not for standard stats
    if stat in myelinated_stats:
        myelinated = True
    elif stat in unmyelinated_stats:
        myelinated = False
    # load the stat
    diameters, presence = load_stat(stat)
    # chose min max values for the axon diameters
    if dmin is None:
        dmin = min(diameters)
    if dmax is None:
        bin_size = diameters[1] - diameters[0]
        dmax = max(diameters) + bin_size
    # perform stat fitting and create generator
    if myelinated and dmax > 10 and not one_gamma:
        popt1, pcov1 = curve_fit(
            two_Gamma,
            xdata=diameters,
            ydata=presence,
            bounds=(
                [1, 2, 0, 1, 0, 0, 2],
                [np.inf, 15, np.inf, np.inf, 15, np.inf, 14],
            ),
        )
        generator = nerve_gen_two_gamma()
        generator.set_bounds(dmin, dmax)
        generator.configure(*popt1)
    else:
        popt1, pcov1 = curve_fit(
            one_Gamma,
            xdata=diameters,
            ydata=presence,
            bounds=([1, 0, 0], [np.inf, 10, np.inf]),
        )
        generator = nerve_gen_one_gamma()
        generator.set_bounds(dmin, dmax)
        generator.configure(*popt1)
    return generator, popt1, pcov1


def create_axon_population(
    N, percent_unmyel=0.7, M_stat="Schellens_1", U_stat="Ochoa_U"
):
    """
    Create a virtual population of axons (no Neuron implementation, axon class not called) of a controled number, with controlled statistics.

    Parameters
    ----------
    N               : int
        Number of axon to generate for the population (Unmyelinated and myelinated)
    percent_unmyel  : float
        ratio of unmyelinated axons in the population. Should be between 0 and 1.
    M_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
    U_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition

    Returns
    -------
    axons_diameters     : np.array
        Array of length N, containing all the diameters of the generated axon population
    axons_type          : np.array
        Array of length N, containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    # create generators
    M_gen, M_popt, M_pcov = create_generator_from_stat(M_stat)
    U_gen, U_popt, U_pcov = create_generator_from_stat(U_stat)
    # number of myelinated and unmyelinated axons to create
    U = int(N * percent_unmyel)
    M = N - U
    pass_info(
        "On "
        + str(N)
        + " axons to generate, there are "
        + str(M)
        + " Myelinated and "
        + str(U)
        + " Unmyelinated"
    )
    # generate the myelinated axons
    xspace1 = np.linspace(1, 20, num=500)
    if len(M_popt) < 4:
        data = one_Gamma(xspace1, *M_popt)
    else:
        data = two_Gamma(xspace1, *M_popt)
    data = data / np.sum(data)
    M_diam_list = np.random.choice(xspace1, M, p=data)
    M_type = np.ones(M)

    # generate the unmyelinated axons
    xspace1 = np.linspace(0.1, 3, num=500)
    data = one_Gamma(xspace1, *U_popt)
    data = data / np.sum(data)
    U_diam_list = np.random.choice(xspace1, U, p=data)
    U_type = np.zeros(U)

    shuffle_mask = np.random.permutation(N)
    axons_diameters = np.concatenate((M_diam_list, U_diam_list))
    axons_type = np.concatenate((M_type, U_type))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type, M_diam_list, U_diam_list


def fill_area_with_axons(
    A, percent_unmyel=0.7, FVF=0.55, M_stat="Schellens_1", U_stat="Ochoa_U"
):
    """
    Create a virtual population of axons (no Neuron implementation, axon class not called) to fill a specified area, with controlled statistics.

    Parameters
    ----------
    A               : float
        surface area to fill, in um**2
    percent_unmyel  : float
        ratio of unmyelinated axons in the population. Should be between 0 and 1.
    FVF             :
        Fiber Volume Fraction estimated for the area. By default set to 0.55
    M_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
    U_stat          : str
        name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition

    Returns
    -------
    axons_diameters     : np.array
        Array  containing all the diameters of the generated axon population
    axons_type          : np.array
        Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    # create generators
    M_gen, M_popt, M_pcov = create_generator_from_stat(M_stat)
    M_diam_list = []
    M_max_diam_value = M_gen.b - 0.01 * (M_gen.b - M_gen.a)
    U_gen, U_popt, U_pcov = create_generator_from_stat(U_stat)
    U_diam_list = []
    U_max_diam_value = U_gen.b - 0.01 * (U_gen.b - U_gen.a)
    # area
    A_ax = 0
    A_total = 0
    # axon lists
    M_diam_list = []
    U_diam_list = []
    # loop
    while A_total < A:
        filled = A_total / A
        progression_popup(
            round(filled * 100), 100, begin_message="\t Area filled at ", endl="\r"
        )
        # print('\t Area filled at ' + f"{round(filled*100)}" + ' percent', end="\r")
        U_or_M = np.random.uniform(0, 1)
        if U_or_M < percent_unmyel:
            # generate a unmyelinated axon
            prov_diam = U_gen.rvs(1, size=1)[0]
            while prov_diam > U_max_diam_value:
                prov_diam = U_gen.rvs(1, size=1)[0]
            U_diam_list.append(prov_diam)
            A_ax += np.power(U_diam_list[-1] / 2, 2) * np.pi
        else:
            prov_diam = M_gen.rvs(1, size=1)[0]
            while prov_diam > M_max_diam_value:
                prov_diam = M_gen.rvs(1, size=1)[0]
            M_diam_list.append(prov_diam)
            A_ax += np.power(M_diam_list[-1] / 2, 2) * np.pi
        A_total = A_ax / FVF
    # concatenate and shuffle
    M = len(M_diam_list)
    U = len(U_diam_list)
    N = M + U
    M_type = np.ones(M)
    U_type = np.zeros(U)
    shuffle_mask = np.random.permutation(N)
    axons_diameters = np.concatenate((M_diam_list, U_diam_list))
    axons_type = np.concatenate((M_type, U_type))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type, M_diam_list, U_diam_list


def shuffle_population(axons_diameters, axons_type):
    """
    Shuffle an axonal population

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)

    Returns
    -------
    axons_diameters : np.array
        shuffled axons diamters
    axons_type      : np.array
        corrresponding axons type
    """
    shuffle_mask = np.random.permutation(len(axons_diameters))
    axons_diameters = axons_diameters[shuffle_mask]
    axons_type = axons_type[shuffle_mask]
    return axons_diameters, axons_type


#############################################
#############################################
#############################################
def init_packing( 
        diam: np.array,
        delta: np.float32,
        y_gc: np.float32,             
        z_gc: np.float32,
        ) -> np.array:
    """
    Initialize Axon packing Algorithm - Internal use Only
    """
    Naxon = len(diam)
    ids = np.arange(Naxon)         #get IDs of each axons
    max_diam = np.max(diam) # get max axon diameter

    #Vector of initial velocity
    velocity= np.zeros([2,Naxon])

    ### create an initial square grid with the population of axons
    N_init = np.int32(np.ceil(np.sqrt(Naxon)) ** 2)
    N_side = np.int32(np.sqrt(N_init))
    size_init = np.sqrt(N_init * (max_diam + delta) ** 2)
    grid_size = (size_init / 2) - (size_init / (2 * N_side))
    grid = np.linspace(-grid_size,grid_size,num=N_side,endpoint=True)
    y_axons = np.tile(grid, N_side) + y_gc
    z_axons = np.repeat(grid, N_side) + z_gc

    ### remove unwanted places
    N_remove = N_init - Naxon
    ind_to_delete = np.random.choice(len(y_axons), size=N_remove, replace=False)
    y_axons = np.delete(y_axons, ind_to_delete)
    z_axons = np.delete(z_axons, ind_to_delete)

    #Define position vector:
    pos = np.array([y_axons, z_axons])

    #Position of center of gravity
    gc = np.array([y_gc,z_gc])

    return(pos,velocity,gc,ids)

def get_axon_combinaisons(ids: np.array)->np.array:
    """
    get all possible combinaison of axons IDs = n(n-1)/2 - Internal use only
    """
    return(np.asarray(list(combinations(ids,2))))

def get_axon_other_id(ids: np.array)->np.array:
    """
    For each axon, get ids all other axon: N*(N-1) - Internal use only
    """
    Nax = len(ids)
    return(np.asarray(list(combinations(ids,Nax-1))))

"""def get_axon_interdistance(
        pos: np.array,
        ids_pairs: np.array
    )-> np.array:
    
    Compute the distance between each pair of axons - Internal use only
    @Note: The sqrt could probably be omitted  to increase speed
    
    #get the pairs of y and z positions of all axons
    y_pairs = np.array([pos[0][ids_pairs[:,0]], pos[0][ids_pairs[:,1]]]).T  #we get the y,z position of each combinaison
    z_pairs = np.array([pos[1][ids_pairs[:,0]], pos[1][ids_pairs[:,1]]]).T
    #we get the dy and dz combinaison 
    dy_pairs = np.diff(y_pairs, axis=1).ravel()
    dz_pairs = np.diff(z_pairs, axis=1).ravel()
    #We return the position.
    return(np.sqrt(dz_pairs**2 + dy_pairs**2))"""

def get_axon_inter_radius(
        diam: np.array,
        ids_pairs:  np.array
    )-> np.array:
    """
    return the inter-radis of each pair of axons - Internal use only
    """
    diam_pair = np.array([diam[ids_pairs[:,0]], diam[ids_pairs[:,1]]]).T  #we get the diameter of each combinaison
    return(np.abs(np.sum(diam_pair, axis=1).ravel())/2)

def get_delta_pairs(pos:np.array, ids_pairs:np.array)-> np.array:
    """
    Evaluate the (z,y) distance of each axons pairs
    """
    return np.diff(np.array([pos[ids_pairs[:,0]], pos[ids_pairs[:,1]]]).T, axis=1).ravel()

def get_deltad_pairs(pos:np.array, ids_pairs:np.array)-> np.array:
    """
    Evaluate the eucledian distance coordinate of each axons pairs - Internal use only
    @Note: The sqrt could probably be omitted to increase speed
    """
    return np.sqrt(get_delta_pairs(pos[0], ids_pairs)**2 + get_delta_pairs(pos[1], ids_pairs)**2)

def get_gravity_dr(pos:np.array,gc:np.array,Naxons:np.int32) -> np.array:
    """
    Evaluate the (z,y) distance to the gravity center of each axons - Internal use only
    """
    return((np.ones([2,Naxons]).T * gc).T - pos)     

def get_gravity_dist(pos:np.array,gc:np.array,Naxons:np.int32) -> np.array:
    """
    Evaluate the eucledian distance to the gravity center of each axons - Internal use only
    @Note: The sqrt could probably be omitted to increase speed
    """
    return(np.sqrt(np.sum((get_gravity_dr(pos,gc,Naxons).T - gc) ** 2,axis = 1)+1e-10))   

def compute_attraction_v(pos:np.array,gc:np.array,Naxons:np.int32,v_att:np.float32)-> np.array:
    """
    Evaluate the attraction velocity for every axons - Internal use only
    """
    return(v_att*get_gravity_dr(pos,gc,Naxons)/get_gravity_dist(pos,gc,Naxons))

def compute_repulsion_v(pos1:np.array, pos2:np.array,v_rep:np.float32)-> np.array:
    """
    Evaluate the repulsion velocity for the colliding axons only - Internal use only
    """
    vnew = v_rep* (pos1-pos2)                        
    return vnew, -vnew

def get_colliding_ids(pos:np.array,id_pairs:np.array,diam_pairs:np.array,delta:np.float32)-> np.array:
    """ 
    get ids of colliding axons - Internal use only
    """
    return(id_pairs[get_deltad_pairs(pos, id_pairs) < (diam_pairs+delta)])

def get_distance_in_range(pos:np.array,id_others:np.array,diam:np.array,delta:np.float32)-> np.array:

    id = np.arange(len(pos[0]))
    id = np.flip(id)

    r_comb = (diam[id_others] + diam[id][:,None])/2

    y_dis = pos[0][id][:,None] - pos[0][id_others] 
    x_dis = pos[1][id][:,None] - pos[1][id_others]
    dist = np.sqrt(y_dis**2 + x_dis**2) - r_comb

    stop_c = 1.2*delta
    if stop_c < 1:
        stop_c = 1
    #low_bound = np.min(dist,axis = 1)>0.95*delta
    #up_bound = np.min(dist,axis = 1)<2*delta
    #print(np.min(dist,axis = 1))
    #exit()
    return(np.max(np.min(dist,axis = 1))<stop_c)



    #print(len(id_others[up_bound]))
    #exit()

    #if len(id_others[low_bound & up_bound]) == len(id):    
    #if len(id_others[up_bound]) == len(id):     
    #    return True
    #else:
    #    return False


def update_axon_packing(pos:np.array,
                        id_pairs:np.array,
                        diam_pairs:np.array,
                        gc: np.array,
                        v_att:np.float32,
                        v_rep:np.float32,
                        delta:np.float32,
                        Naxon:np.int32)->np.array:
    """ 
    Update the axon array position by 1 increment - Internal use only
    """

    velocity = compute_attraction_v(pos,gc,Naxon,v_att)
    ic = get_colliding_ids(pos,id_pairs,diam_pairs,delta)
    velocity[:,ic[:,0]], velocity[:,ic[:,1]] = compute_repulsion_v(pos[:,ic[:,0]], pos[:,ic[:,1]],v_rep)
    return (pos + velocity)

def axon_packer(diameters: np.array,
                 y_gc:np.float32 = 0,
                 z_gc:np.float32 = 0,
                 delta: np.float32 = 0.5,
                 n_iter: np.int32 = 20000,
                 v_att: np.float32 = 0.01,
                 v_rep: np.float32 = 0.1,
                 monitor = False,
                 monitoring_Folder="",
                 n_monitor = 200):
    """
    Axon Packing algorithm: this operation takes a vector of diameter (random population) and places it at best. The used algorithm is largely based on [1]

    Parameters
    ----------
    diameters           : np.array
        Array containing all the diameters of the axon population to pack
    delta               : float
        minimal inter-axon distance to respect before considering collision, in um
    n_iter            : int
        Number of iterations
    v_att               : float
        vector norm for attraction velocity, in um per iteration
    v_rep               : float
        vector norm for repulsion velocity, in um per iteration
    y_gc                : float
        y coordinate of the gravity center for the packing, in um
    z_gc                : float
        z coordinate of the gravity center for the packing, in um
    monitor             : bool
        monitor the packing algorithm by saving regularly plot of the population
    monitoring_Folder   : str
        where to save the monitoring plots
    n_monitor           : int
        number of iterration between two successive plots when monitoring the algorithm

    Returns
    -------
    y_axons         : np.array
        Array containing the y coordinates of axons, in um
    z_axons         : np.array
        Array containing the z coordinates of axons, in um

    Note
    ----
    - scientific reference
        [1] Mingasson, T., Duval, T., Stikov, N., and Cohen-Adad, J. (2017). AxonPacking: an open-source software to simulate arrangements of axons in white matter. Frontiers in neuroinformatics, 11, 5.
    - This algorithm cannot be parallelized for the moment.
    - dev Note
        the algorithm perform a fixed number of iteration, the code could evoluate to tke into account the FVF convergence, however this value depends on the range on number of axons.
    """
    pass_info("Axon packing initiated. This might take a while...")
    pos,velocity,gc,ids= init_packing(diameters,delta,y_gc,z_gc)
    max_pos = 2.2 * (np.max(pos[0]) - y_gc)
    id_pairs = get_axon_combinaisons(ids)
    diam_pair = get_axon_inter_radius(diameters,id_pairs)
    Naxon = len(pos[0])

    id_others = get_axon_other_id(ids)

    #for _ in track(range (n_iter)):
    #    pos = update_axon_packing(pos,id_pairs,diam_pair,gc,v_att,v_rep,delta,Naxon)
    if monitor:
        fig, ax = plt.subplots(figsize=(8, 8))
        ax.set_axis_off()
        fig.add_axes(ax)
    for i in track(range (n_iter)):
        pos = update_axon_packing(pos,id_pairs,diam_pair,gc,v_att,v_rep,delta,Naxon)
        
        if monitor and i %n_monitor == 0:
            y_prov = pos[0].copy()
            z_prov = pos[1].copy()
            ax.cla()
            plot_population(diameters, y_prov, z_prov,ax,max_pos, y_gc=y_gc, z_gc=z_gc)
            plt.savefig(monitoring_Folder+"vignette_"+str(i)+".png")
        
        '''
        if (get_distance_in_range(pos,id_others,diameters,delta)):
        #t.append(get_distance_in_range(pos,id_others,diameters,delta))
            pass_info("Stop criterion reached!")
            break
        '''
    y_axons = pos[0].copy()
    z_axons = pos[1].copy()
    pass_info("Packing done!")
    y_c, z_c = get_barycenter(diameters,y_axons,z_axons)    #center the population back to 0,0
    return y_axons-y_c, z_axons-z_c

def expand_pop(y_axons:np.array, z_axons:np.array, factor:float) -> np.array:
    """
    Expand population of placed axons by a specified number

    Parameters
    ----------
    y_axons     : np.array
        y coordinate of the axons, in um
    z_axons     : np.array
        z coordinate of the axons, in um
    factor          : float
        expansion factor, unitless
    """
    if (factor<1):
        rise_warning("expansion factor must be greater than one. Factor set to 1")
        factor = 1
    return(y_axons*factor,z_axons*factor)

def remove_collision(axons_diameters:np.array,y_axons:np.array,z_axons:np.array, axon_type:np.array, delta: float=0)-> np.array:
    """
    Remove collinding axons in a population

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    axon_type   : np.array
        type of the axon (Myelinated = 1; Unmyelinated = 0)
    delta               : float
        minimal inter-axon distance to respect before considering collision, in um
    """

    Naxon = len(axons_diameters)
    ids = np.arange(Naxon)    
    id_pairs = get_axon_combinaisons(ids)
    pos = np.zeros([2,Naxon])
    pos[0,:] = y_axons
    pos[1,:] = z_axons
    diam_pairs = get_axon_inter_radius(axons_diameters,id_pairs)
    ic = get_colliding_ids(pos,id_pairs,diam_pairs,delta) 
    ic = ic[:,0]
    if len(ic)>0:
        rise_warning(f"{len(ic)} axon collisions detected - Axons discarded.")
    return(np.delete(axons_diameters,ic),np.delete(y_axons,ic),np.delete(z_axons,ic),np.delete(axon_type,ic))

def get_circular_contour(axons_diameters:np.array,y_axons:np.array,z_axons:np.array, delta: float = 10)-> float:
    """
    Get a circular contour diameter of the axon population

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    delta               : float
        distance between the contour and the closest axon, in um
    """
    dist_axon = np.sqrt((y_axons ** 2 + z_axons**2))
    dist_axon+= axons_diameters/2
    radius = np.max(dist_axon) + delta
    return(radius*2)

def remove_outlier_axons(axons_diameters:np.array, y_axons:np.array, z_axons:np.array, axon_type:np.array, diameter: float = 10)-> np.array:
    """
    Remove axons in a population located outside a circular border, defined by its diameter

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    axon_type   : np.array
        type of the axon (Myelinated = 1; Unmyelinated = 0)
    diameter               : float
        diameter of the circular border, in um
    """
    dist_axon = np.sqrt((y_axons ** 2 + z_axons**2))
    dist_axon+= axons_diameters/2
    inside_border = dist_axon <= diameter/2
    n_remove = len(np.where(inside_border==False)[0])
    if (n_remove>0):
        rise_warning(f"{n_remove} outlier axons discarded.")

    return(axons_diameters[inside_border],y_axons[inside_border],z_axons[inside_border],axon_type[inside_border])

def get_barycenter(axons_diameters, y_axons, z_axons):
    """
    Compute barycenter of the population

    Parameters
    ----------
    axons_diameters : np.array
        array containing the axons diameters
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    """
    diam_sum = np.sum(axons_diameters)
    return(np.sum(axons_diameters * y_axons)/diam_sum,np.sum(axons_diameters * z_axons)/diam_sum)

def plot_population(diameters, y_axons, z_axons,ax,size, axon_type = None, y_gc=0, z_gc=0)->None:
    """
    Display a population of axons. 

    Parameters
    ----------
    diameters   : np.array
        diamters of the axons to display, in um
    y_axons     : np.array
        y coordinate of the axons to display, in um
    z_axons     : np.array
        z coordinate of the axons to display, in um
    ax          : matplotlib.axes
        (sub-) plot of the matplotlib figure
    size        : float
        size of the window as a square side, in um
    axon_type   : np.array
        type of the axon (Myelinated = 1; Unmyelinated = 0) - Optionnal
    title       : str
        title of the figure
    y_gc        : float
        y coordinate of the gravity, in um
    z_gc        : float
        z coordinate of the gravity, in um
    """
    if (axon_type is None):
        for k in range(len(diameters)):
            ax.add_patch(plt.Circle((y_axons[k], z_axons[k]), diameters[k] / 2, color="r", fill=True))
    else:
        # plot the final results, with distinction between un- and myelinated axons
        myelinated_mask = np.argwhere(axon_type == 1)
        y_myelinated = y_axons[myelinated_mask]
        z_myelinated = z_axons[myelinated_mask]
        M_diam_list = diameters[myelinated_mask]
        for k in range(len(y_myelinated)):
                ax.add_patch(plt.Circle((y_myelinated[k], z_myelinated[k]), M_diam_list[k]/2, color='r',fill=True))


        unmyelinated_mask = np.argwhere(axon_type == 0) 
        y_unmyelinated = y_axons[unmyelinated_mask]
        z_unmyelinated = z_axons[unmyelinated_mask]
        U_diam_list = diameters[unmyelinated_mask]
        for k in range(len(y_unmyelinated)):
                ax.add_patch(plt.Circle((y_unmyelinated[k], z_unmyelinated[k]), U_diam_list[k]/2, color='b',fill=True))

    ax.set_xlim(-size / 2 + y_gc, size / 2 + y_gc)
    ax.set_ylim(-size / 2 + z_gc, size / 2 + z_gc)


def save_axon_population(
    f_name, axons_diameters, axons_type, y_axons=None, z_axons=None, comment=None
):
    """
    Save a placed axonal population to a file

    Parameters
    ----------
    f_name          : str
        name of the file to store the population
    axons_diameters : np.array
        array containing the axons diameters
    axons_type      : np.array
        array containing the axons type ('1' for myelinated, '0' for unmyelinated)
    y_axons     : np.array
        y coordinate of the axons to store, in um
    z_axons     : np.array
        z coordinate of the axons to store, in um
    comment         : str
        comment added in the header of the file, optional
    """
    if y_axons is None:
        y_axons = np.zeros([len(axons_diameters)])
        y_axons[:] = np.nan

    if z_axons is None:
        z_axons = np.zeros([len(axons_diameters)])
        z_axons[:] = np.nan

    f = open(f_name, "w")
    if comment is not None:
        line = "# " + comment
        f.write(line)
    for k in range(len(axons_diameters)):
        line = (
            str(axons_diameters[k])
            + ", "
            + str(axons_type[k])
            + ", "
            + str(y_axons[k])
            + ", "
            + str(z_axons[k])
            + "\n"
        )
        f.write(line)
    f.close()


def load_axon_population(f_name):
    """
    Load a placed population from a file

    Parameters
    ----------
    f_name          : str
        name of the file where the population is stored

    Returns
    -------
    axons_diameters     : np.array
        Array containing all the diameters of the generated axon population
    axons_type          : np.array
        Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
    y_axons     : np.array
        y coordinate of the axons, in um
    z_axons     : np.array
        z coordinate of the axons, in um
    M_diam_list         : np.array
        list of myelinated only diameters
    U_diam_list         : np.array
        list of unmyelinated only diamters
    """
    population_file = np.genfromtxt(f_name, delimiter=",", comments="#")
    axons_diameters = population_file[:, 0]
    axons_type = population_file[:, 1]
    y_axons = population_file[:, 2]
    z_axons = population_file[:, 3]
    ind_myel = np.argwhere(axons_type == 1)
    ind_unmyel = np.argwhere(axons_type == 0)
    M_diam_list = axons_diameters[ind_myel]
    U_diam_list = axons_diameters[ind_unmyel]
    

    if (np.isnan(y_axons).all() or np.isnan(z_axons).all()):
        rise_warning("Loaded population has no y,z coordinates.")

    return axons_diameters, axons_type, M_diam_list, U_diam_list, y_axons, z_axons



"""
Module containing a top class from which all other NRV classes inherit.
This is mainly used to create generic methods such as save and load.
"""

from abc import ABCMeta, abstractmethod
from copy import deepcopy

# sys used in an eval
import sys
import numpy as np
from numpy import iterable

from ._file_handler import json_dump, json_load
from ._log_interface import pass_debug_info

########################################
#           check object               #
########################################


def is_NRV_class(x):
    """
    Check if the object x is a ``NRV_class``.

    Parameters
    ----------
    x   : any
        object to check.

    Returns
    -------
    bool
    """
    return isinstance(x, NRV_class)


def is_NRV_class_list(x):
    """
    check if the object x is a list containing only ``NRV_class``.

    Parameters
    ----------
    x   : any
        object to check.

    Returns
    -------
    bool
    """
    if iterable(x):
        for xi in x:
            if not is_NRV_class(xi):
                return False
        return True
    return False


def is_NRV_class_dict(x):
    """
    check if the object x is a dictionary containing only ``NRV_class``.

    Parameters
    ----------
    x   : any
        object to check.

    Returns
    -------
    bool
    """
    if isinstance(x, dict):
        for xi in x.values():
            if not is_NRV_class(xi):
                return False
        return True
    return False


##########################################
#           check dictionaries           #
##########################################
def is_NRV_dict(x):
    """
    Check if the object x is a dictionary of saved ``NRV_class``.

    Parameters
    ----------
    x : any
        object to check.

    Returns
    -------
    bool
    """
    if isinstance(x, dict):
        if "nrv_type" in x:
            return True
    return False


def is_NRV_dict_list(x):
    """
    Check if the object x is a list of dictionary of saved ``NRV_class``.

    Parameters
    ----------
    x : any
        object to check.

    Returns
    -------
    bool
    """
    if iterable(x):
        if len(x) > 0:
            for xi in x:
                if not (is_NRV_dict(xi)):
                    return False
            return True
    return False


def is_NRV_dict_dict(x):
    """
    Check if the object x is a dictionary containing dictionaries of saved ``NRV_class``.

    Parameters
    ----------
    x : any
        object to check.

    Returns
    -------
    bool
    """
    if isinstance(x, dict):
        for key in x:
            if not (is_NRV_dict(x[key])):
                return False
        return True
    return False


def is_NRV_object_dict(x):
    """
    Check if the object x is_NRV_dict, is_NRV_dict_list or is_NRV_dict_dict.

    Parameters
    ----------
    x : any
        object to check.

    Returns
    -------
    bool
    """
    return is_NRV_dict(x) or is_NRV_dict_list(x) or is_NRV_dict_dict(x)


######################################
#       numpy compatibility          #
######################################
def is_empty_iterable(x):
    """
    check if the object x is an empty iterable

    Parameters
    ----------
    x : any
        object to check.

    Returns
    -------
    bool
    """
    if not np.iterable(x):
        return False
    if len(x) == 0:
        return True
    return False


######################################
#            NRV Class               #
######################################


class NRV_class(metaclass=ABCMeta):
    """
    Instanciate a basic NRV class
    NRV Class are empty shells, defined as abstract classes of which every class in NRV
    should inherite. This enable automatic context backup with save and load methods.
    """

    @abstractmethod
    def __init__(self):
        """
        Init method for ``NRV_class``
        """
        self.__NRVObject__ = True
        self.nrv_type = self.__class__.__name__
        pass_debug_info(self.nrv_type, " initialized")

    def __del__(self):
        """
        Destructor for ``NRV_class``
        """
        pass_debug_info(self.nrv_type, " deleted")
        keys = list(self.__dict__.keys())
        for key in keys:
            del self.__dict__[key]

    def save(self, save=False, fname="nrv_save.json", blacklist=[], **kwargs) -> dict:
        """
        Generic saving method for ``NRV_class`` instance.

        Parameters
        ----------
        save: bool, optional
            If True, saves the NRV object in a json file, by default False.
        fname : str, optional
            Name of the json file
        blacklist : dict, optional
            Dictionary containing the keys to be excluded from the saving.
        **kwargs : dict, optional
            Additional arguments to pass to the ``save`` method of the NRV object.

        Returns
        -------
        key_dict : dict
            dictionary containing the original instance data in a `jsonisable` format.

        Note
        ----
        - This ``save`` method does not save the object to a `json` file by default: It only\
            returns a dictionary containing the original instance data in a jsonisable format.\
            However, this is the simplest way to do it by setting the ``save`` parameter to ``True``.
        - The dictionary returned by this `save` method can be modified without having any impact on the\
            the original instance (the items are deep copies of the instance's attributes).
        """
        key_dic = {}
        for key in self.__dict__:
            if key not in blacklist:
                if is_NRV_class(self.__dict__[key]):
                    key_dic[key] = self.__dict__[key].save(**kwargs)
                elif is_NRV_class_list(self.__dict__[key]):
                    key_dic[key] = []
                    for i in range(len(self.__dict__[key])):
                        key_dic[key] += [self.__dict__[key][i].save(**kwargs)]
                elif is_NRV_class_dict(self.__dict__[key]):
                    key_dic[key] = {}
                    for i in self.__dict__[key]:
                        key_dic[key][i] = self.__dict__[key][i].save(**kwargs)
                else:
                    key_dic[key] = deepcopy(self.__dict__[key])
        if save:
            json_dump(key_dic, fname)
        return key_dic

    def load(self, data, blacklist={}, **kwargs) -> None:
        """
        Generic loading method for ``NRV_class`` instance

        Parameters
        ----------
        data : str, dict, nrv_class
            data from which the object should be generated:

                - if str, data will be loaded from the corresponding json file
                - if dict, data will be loaded from a dictionnary
                - if nrv_class, same object will be returned
        blacklist : dict, optional
            Dictionary containing the keys to be excluded from the load
        **kwargs : dict, optional
            Additional arguments to be passed to the load method of the NRV object
        """
        if isinstance(data, str):
            key_dic = json_load(data)
        else:
            key_dic = data
        for key in self.__dict__:
            if key in key_dic and key not in blacklist:
                if is_NRV_object_dict(key_dic[key]):
                    self.__dict__[key] = load_any(key_dic[key], **kwargs)
                elif isinstance(self.__dict__[key], np.ndarray):
                    self.__dict__[key] = np.array(key_dic[key])
                elif isinstance(self.__dict__[key], set):
                    self.__dict__[key] = set(key_dic[key])
                elif is_empty_iterable(key_dic[key]):
                    self.__dict__[key] = eval(self.__dict__[key].__class__.__name__)()
                else:
                    self.__dict__[key] = key_dic[key]

    def set_parameters(self, **kawrgs) -> None:
        """
        Generic method to set any attribute of ``NRV_class`` instance

        Parameters
        ----------
        ***kwargs
            Key arguments containing one or multiple parameters to set.

        Examples
        --------
        As the :class:`~nrv.nmod.myelinated.myelinated` inherits from NRV_class-class parameters, such as diameter and lenght can be set with `set_parameters`.

        >>> ax = nrv.myelinated()
        >>> print(ax.d, ax.L)
        10, 10000
        >>> ax.set_parameters(d=6, L=1000)
        >>> print(ax.d, ax.L)
        6, 1000
        """
        for key in kawrgs:
            if key in self.__dict__:
                self.__dict__[key] = kawrgs[key]

    def get_parameters(self):
        """
        Generic method returning all the atributes of an NRV_class instance

        Returns
        -------
            dict : dictionnary of all atributes of ``NRV_class`` instance
        """
        return self.__dict__


def load_any(data, **kwargs)->NRV_class:
    """
    loads any type of NRV object from a json file or a dictionary generated with NRV_class.save

    Parameters
    ----------
    data : str, dict, nrv_class
        data from which the object should be generated:

            - if str, data will be loaded from the corresponding json file
            - if dict, data will be loaded from a dictionnary
            - if nrv_class, same object will be returned
    **kwargs
        additionnal argument to use for the loading

    Returns
    -------
    nrv_obj: any (NRV_class)


    """
    if isinstance(data, str):
        key_dic = json_load(data)
    else:
        key_dic = data
    # test if NRV class
    if is_NRV_class(key_dic) or is_NRV_class_list(key_dic):
        nrv_obj = key_dic
    # test if NRV dict
    elif is_NRV_dict(key_dic):
        nrv_type = key_dic["nrv_type"]
        nrv_obj = eval('sys.modules["nrv"].' + nrv_type)()
        nrv_obj.load(key_dic, **kwargs)
    elif is_NRV_dict_dict(key_dic):
        nrv_obj = {}
        for key in key_dic:
            nrv_obj[key] = load_any(key_dic[key], **kwargs)
    elif is_NRV_dict_list(key_dic):
        nrv_obj = []
        for i in key_dic:
            nrv_obj += [load_any(i, **kwargs)]
    else:
        nrv_obj = key_dic
    return nrv_obj



import numpy as np
from ...backend._NRV_Class import NRV_class, load_any
from ...backend._NRV_Simulable import NRV_simulable
from ...utils._stimulus import stimulus


class context_modifier(NRV_class):
    """
    Instantiate a context modifier: Callable object which modify a static context, regarding a vector of tunning parameters, to generate a new local context

    Notes
    -----
    `context_modifier` is an abstract class used to inherit new context modifier.

    Parameters
    ----------
    extracel_context        : bool
        specifies whether to load extracel_context with the static context
    intracel_context        : bool
        specifies whether to load intracel_context with the static context
    rec_context     : bool
        specifies whether to load rec_context with the static context
    """

    def __init__(
        self, extracel_context=True, intracel_context=False, rec_context=False
    ):
        super().__init__()
        self.extracel_context = extracel_context
        self.intracel_context = intracel_context
        self.rec_context = rec_context

    def __call__(
        self, X: np.ndarray, static_context: NRV_simulable, **kwargs
    ) -> NRV_simulable:
        """
        Modify ``static_context`` tunning paramters vector ``X``.

        Parameters
        ----------
        X               : np.array (X,)
            Vector of that will lead to the modification of the static context
        static_context  : NRV_simulable
            Static context which should be modify to optain th local context

        Returns
        -------
        local_context  : NRV_simulable
            simulable object generated from the vector X and the static_context

        """
        self.set_parameters(**kwargs)
        return load_any(
            static_context,
            extracel_context=self.extracel_context,
            intracel_context=self.intracel_context,
            rec_context=self.rec_context,
        )


class stimulus_CM(context_modifier):
    """
    Generic context modifier which generate a stimulus from the input tuning parameters.

    Note
    ----
    When an instance of this class is called the three following steps are done:
        1. The input vector is interpolated with the `interpolator`. If not set, the interpolated vector will be the equal to the input vector.
        2. The stimulus is generated from interpolated values with  `stim_gen` function. If not set, interpolated values are set for the stimulus amplitude at constant sample rate (from zero to an argument ``t_sim``)
        3. The generated stimulus is added to an electrode of a static context.

    Parameters
    ----------
    stim_ID                 : int, optional
        ID of the electrode which should be adapted, by default 0
    interpolator            : callable, optional
        if not None function use to interpolate the input vector, by default None
    intrep_kwargs           : dict, optional
        key arguments to use with the `interpolator`, by default {}
    stim_gen                : callable, optional
        function use to generate the stimulus from the interpolated values, by default None
    stim_gen_kwargs         : dict, optional
        key arguments used for the , by default {}
    extracel_context        : bool
        specifies whether to load extracel_context with the static context
    intracel_context        : bool
        specifies whether to load intracel_context with the static context
    rec_context             : bool
        specifies whether to load rec_context with the static context
    """

    def __init__(
        self,
        stim_ID: int = 0,
        interpolator: callable = None,
        intrep_kwargs: dict = {},
        stim_gen: callable = None,
        stim_gen_kwargs: dict = {},
        extracel_context: bool = True,
        intracel_context: bool = False,
        rec_context: bool = False,
        **kwargs,
    ):
        super().__init__(
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )
        self.stim_ID = stim_ID
        self.interpolator = interpolator
        self.intrep_kwargs = intrep_kwargs
        self.stim_gen = stim_gen
        self.stim_gen_kwargs = stim_gen_kwargs
        for key in kwargs:
            self.intrep_kwargs[key] = kwargs[key]
            self.stim_gen_kwargs[key] = kwargs[key]

    def interpolate(self, X: np.ndarray) -> np.ndarray:
        """
        interpolate the input vector with the `interpolator`. If not set, the interpolated vector will be the equal to the input vector.
        Parameters
        ----------
        X : np.ndarray
            Input vector to interpolate.

        Returns
        -------
        np.ndarray
            Interpolated vector.
        """
        if self.interpolator is not None:
            X_interp = self.interpolator(X, **self.intrep_kwargs)
        else:
            X_interp = X
        return X_interp

    def stimulus_generator(self, X_interp: np.ndarray) -> stimulus:
        """
        generated from interpolated values with  `stim_gen` function. If not set, interpolated values are set for the stimulus amplitude at constant sample rate (from zero to an argument ``t_sim``)

        Parameters
        ----------
        X_interp : np.ndarray
            Interpolated values.

        Returns
        -------
        stimulus
            stimulus generated from the values
        """
        if self.stim_gen is not None:
            stim = self.stim_gen(X_interp, **self.stim_gen_kwargs)
        else:
            t_sim = self.stim_gen_kwargs["t_sim"]
            stim = stimulus()
            N = len(X_interp)
            stim.s = X_interp
            stim.t = np.linspace(0, t_sim, N)
        return stim

    def __update_t_sim(self, local_sim):
        if "t_sim" not in self.stim_gen_kwargs:
            if "t_sim" in self.intrep_kwargs:
                self.stim_gen_kwargs["t_sim"] = self.intrep_kwargs["t_sim"]
            else:
                self.stim_gen_kwargs["t_sim"] = local_sim.t_sim

    def __call__(
        self, X: np.ndarray, static_sim: NRV_simulable, **kwargs
    ) -> NRV_simulable:
        """
        Modify ``static_context`` tunning paramters vector ``X``.

        Note
        ----
        The ``t_sim`` argument need can be added for `stimulus_generator`. If the instance is called from :class:`~nrv.optim.CostFunctions.cost_function` it will be automatically added from the simulation parameter.

        Parameters
        ----------
        X               : np.array (X,)
            Vector of that will lead to the modification of the static context
        static_context  : NRV_simulable
            Static context which should be modify to optain th local context

        Returns
        -------
        local_context  : NRV_simulable
            simulable object generated from the vector X and the static_context
        """
        local_sim = super().__call__(X, static_sim, **kwargs)
        self.__update_t_sim(local_sim)
        X_inter = self.interpolate(X)
        stim = self.stimulus_generator(X_inter)
        local_sim.extra_stim.change_stimulus_from_electrode(self.stim_ID, stim)
        return local_sim


class biphasic_stimulus_CM(stimulus_CM):
    r"""
    Context modifier which generate a stimulus biphasic stimulus from the input tuning parameters.

    Parameters
    ----------
    stim_ID                 : int, optional
        ID of the electrode which should be adapted, by default 0
    start       : float or str
        starting time of the waveform, in ms
    s_cathod    : float or str
        cathodic (negative stimulation value) current, in uA
    s_anod      : float or str
        anodic (positive stimulation value) current and, in uA
    t_inter     : float or str
        inter pulse timing, in ms
    s_ratio     : float or str
        if s_cathod set to None, s_cathod is set as: :math:`s_{anod}*s_{ratio}`. by default 0.2
    stim_gen                : callable, optional
        function use to generate the stimulus from the interpolated values, by default None
    stim_gen_kwargs         : dict, optional
        key arguments used for the , by default {}
    extracel_context        : bool
        specifies whether to load extracel_context with the static context
    intracel_context        : bool
        specifies whether to load intracel_context with the static context
    rec_context             : bool
        specifies whether to load rec_context with the static context

    WARNING
    -------
    ``s_cathod`` must always positive, the user give here the absolute value

    Note
    ----
    Note that `start`, `s_cathod`, `s_anod` and `t_inter` are the paramters used to generate a :meth:`~nrv.fmod.stimulus.stimulus.harmonic_pulse` stimulus. They can either be:
         - set as ``float``, to a constant value (wich will not be changed by the )
         - set as ``str`` ("1", "2", ..) of the index of the value in the input vector.

    See Also
    --------
    :doc:`tutorials 5</tutorials/5_first_optimization>`: First optimization problem using NRV

    """

    def __init__(
        self,
        stim_ID: int = 0,
        start: float = 1,
        s_cathod: float | str = 100,
        t_cathod: float | str = 60e-3,
        t_inter: float | str = 40e-3,
        s_anod: float | str = None,
        s_ratio: float | str = 0.2,
        stim_gen: callable = None,
        stim_gen_kwargs: dict = {},
        extracel_context: bool = True,
        intracel_context: bool = False,
        rec_context: bool = False,
    ):
        super().__init__(
            stim_ID=stim_ID,
            stim_gen=stim_gen,
            stim_gen_kwargs=stim_gen_kwargs,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )
        self.start = start
        self.s_cathod = s_cathod
        self.t_cathod = t_cathod
        self.t_inter = t_inter
        self.s_anod = s_anod
        self.s_ratio = s_ratio

    def __set_values(self, X: np.ndarray):
        start = self.start
        s_cathod = self.s_cathod
        t_cathod = self.t_cathod
        t_inter = self.t_inter
        if isinstance(self.start, str):
            start = X[int(self.start)]
        if isinstance(self.s_cathod, str):
            s_cathod = X[int(self.s_cathod)]
        if isinstance(self.t_cathod, str):
            t_cathod = X[int(self.t_cathod)]
        if isinstance(self.t_inter, str):
            t_inter = X[int(self.t_inter)]
        s_anod = self.s_anod
        if s_anod is not None:
            if isinstance(self.s_anod, str):
                s_anod = X[int(self.s_anod)]
        else:
            s_ratio = self.s_ratio
            if isinstance(self.s_ratio, str):
                s_ratio = X[int(self.s_ratio)]
            s_anod = s_cathod * s_ratio
        return start, s_cathod, t_cathod, t_inter, s_anod

    def stimulus_generator(self, X_interp) -> stimulus:
        if self.stim_gen is not None:
            stim = self.stim_gen(X_interp, **self.stim_gen_kwargs)
        else:
            stim = stimulus()
            start, s_cathod, t_cathod, t_inter, s_anod = self.__set_values(X_interp)
            stim.biphasic_pulse(start, s_cathod, t_cathod, s_anod, t_inter)
            return stim


class harmonic_stimulus_CM(stimulus_CM):
    """
    Context modifier which generate a stimulus harmonic stimulus from the input tuning parameters.

    Parameters
    ----------
    stim_ID : int, optional
        ID of the electrode which should be adapted, by default 0
    start : float or str
        starting time of the waveform, in ms
    amplitude : float, optional
        _description_, by default 100
    t_pulse : float, optional
        _description_, by default 60e-3
    dt : float, optional
        _description_, by default 0
    stim_gen : callable, optional
        _description_, by default None
    stim_gen_kwargs : dict, optional
        _description_, by default {}
    extracel_context        : bool
        specifies whether to load extracel_context with the static context
    intracel_context        : bool
        specifies whether to load intracel_context with the static context
    rec_context             : bool
        specifies whether to load rec_context with the static context
    """

    def __init__(
        self,
        stim_ID: int = 0,
        start: float = 1,
        amplitude: float = 100,
        t_pulse: float = 60e-3,
        dt: float = 0,
        stim_gen: callable = None,
        stim_gen_kwargs: dict = {},
        extracel_context: bool = True,
        intracel_context: bool = False,
        rec_context: bool = False,
    ):
        super().__init__(
            stim_ID=stim_ID,
            stim_gen=stim_gen,
            stim_gen_kwargs=stim_gen_kwargs,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )
        self.start = start
        self.amplitude = amplitude
        self.t_pulse = t_pulse
        self.dt = dt

    def stimulus_generator(self, X_interp) -> stimulus:
        self.amplitude = X_interp[0]
        N = (len(X_interp) - 1) // 2
        self.amp_list = []
        self.phase_list = []

        for k in range(N):
            self.amp_list.append(X_interp[2 * k + 1])
            self.phase_list.append(X_interp[2 * k + 2])
        stim = stimulus()
        stim.harmonic_pulse(
            start=self.start,
            t_pulse=self.t_pulse,
            amplitude=self.amplitude,
            amp_list=self.amp_list,
            phase_list=self.phase_list,
            dt=self.dt,
        )
        return stim


class harmonic_stimulus_with_pw_CM(stimulus_CM):
    def __init__(
        self,
        stim_ID: int = 0,
        start: float = 1,
        amplitude: float = 100,
        t_pulse: float = 60e-3,
        dt: float = 0,
        stim_gen: callable = None,
        stim_gen_kwargs: dict = {},
        extracel_context: bool = True,
        intracel_context: bool = False,
        rec_context: bool = False,
    ):
        super().__init__(
            stim_ID=stim_ID,
            stim_gen=stim_gen,
            stim_gen_kwargs=stim_gen_kwargs,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )
        self.start = start
        self.amplitude = amplitude
        self.t_pulse = t_pulse
        self.dt = dt

    def stimulus_generator(self, X_interp) -> stimulus:
        self.amplitude = X_interp[0]
        self.t_pulse = X_interp[1]
        N = (len(X_interp) - 2) // 2
        self.amp_list = []
        self.phase_list = []

        for k in range(N):
            self.amp_list.append(X_interp[2 * k + 2])
            self.phase_list.append(X_interp[2 * k + 3])
        stim = stimulus()
        stim.harmonic_pulse(
            start=self.start,
            t_pulse=self.t_pulse,
            amplitude=self.amplitude,
            amp_list=self.amp_list,
            phase_list=self.phase_list,
            dt=self.dt,
        )
        return stim



"""
NRV-:class:`.FEMResults` handling.
"""

import gmsh
import numpy as np
import scipy
from time import perf_counter

from dolfinx.fem import Expression, Function, functionspace
from dolfinx.io.gmshio import model_to_mesh
from dolfinx.io.utils import XDMFFile, VTXWriter, VTKFile
from dolfinx.cpp.mesh import entities_to_geometry
from dolfinx.geometry import (
    bb_tree,
    compute_colliding_cells,
    compute_collisions_points,
    compute_closest_entity,
    compute_distance_gjk,
    create_midpoint_tree,
)
from mpi4py.MPI import COMM_WORLD


from ....backend._file_handler import rmv_ext
from ....backend._log_interface import rise_error, rise_warning
from ....backend._parameters import parameters
from ....backend._NRV_Class import NRV_class
from ..mesh_creator._MshCreator import (
    is_MshCreator,
    clear_gmsh,
)


###############
## Functions ##
###############
def is_sim_res(result):
    """
    check if an object is a FEMResults, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a FEMResults object
    """
    return isinstance(result, FEMResults)


def save_sim_res_list(sim_res_list, fname, dt=1.0):
    r"""
    save a list of SimResults in a .bp folder which can be open with PrarView

    Parameters
    ----------
    sim_res_list : list(FEMResults)
        list of :class:`.FEMResults` to be saved
    fname : str
        File name and path
    ftype : str
        - if True .bp folder, else saved in a .xdmf file.
    dt : float
        time step between each results

    Warning
    -------
    For this function to work dolfinx and ParaView must be up to date:
        -   dolfinx \\(\\geq\\) 0.8.0
        -   ParaView \\(\\geq\\) 5.12.0
    """
    N_list = len(sim_res_list)
    fname = rmv_ext(fname) + ".bp"
    vcp = sim_res_list[0].vout
    vtxf = VTXWriter(sim_res_list[0].domain.comm, fname, vcp)
    for i in range(N_list):
        vcp = sim_res_list[i].vout
        vtxf.write(i * dt)
    vtxf.close()


def read_gmsh(mesh, comm=COMM_WORLD, rank=0, gdim=3):
    """
    Given a mesh_file or a MeshCreator returns mesh cell_tags and facet_tags
    (see dolfinx.io.gmshio.model_to_mesh for more details)
    NB: copy of dolfinx.io.gmshio.read_from_msh verbose from gmsh

    Parameters
    ----------
    mesh_file : str,
        Path and name of mesh file
        NB: extention is not required but must be a .msh file
    comm            : tupple (str, int)
        The MPI communicator to use for mesh creation
    rank            : tupple (str, int)
        The rank the Gmsh model is initialized on
    gdim            : int
        dimension of the mesh, by default 3

    Returns
    -------
    output  :  tuple(3)
        (domain, cell_tag, facet_tag)
    """
    if isinstance(mesh, str):
        mesh_file = rmv_ext(mesh) + ".msh"
        clear_gmsh()
        gmsh.initialize()
        gmsh.option.setNumber("General.Verbosity", 2)
        gmsh.model.add("Mesh from file")
        gmsh.merge(mesh_file)
        output = model_to_mesh(gmsh.model, comm=comm, rank=rank, gdim=gdim)
        gmsh.finalize()
    elif is_MshCreator(mesh):
        output = model_to_mesh(mesh.model, comm=comm, rank=rank, gdim=gdim)
    else:
        rise_error("mesh should be either a filename or a MeshCreator")
    return output


def domain_from_meshfile(mesh_file):
    """
        return only the domain from mesh_file

    Parameters
    ----------
    mesh_file : str
        Path and name of mesh file
        NB: extention is not required but must be a .msh file

    Returns
    -------
        output  :
            domain of the mesh contain in the mesh_file
    """
    return read_gmsh(mesh_file)[0]


def V_from_meshfile(mesh_file, elem=("Lagrange", 1)):
    mesh = domain_from_meshfile(mesh_file)
    V = functionspace(mesh, elem)
    return V


def closest_point_in_mesh(mesh, point, tree, tdim, midpoint_tree):
    points = np.reshape(point, (1, 3))
    entity = compute_closest_entity(tree, midpoint_tree, mesh, points)
    mesh_geom = mesh.geometry.x
    geom_dofs = entities_to_geometry(mesh._cpp_object, tdim, entity, False)
    mesh_nodes = mesh_geom[geom_dofs][0]
    displacement = compute_distance_gjk(points, mesh_nodes)
    return entity, points[0] - displacement


class FEMResults(NRV_class):
    """
    Result of a FEMSimulation.
    Store the resulting function space giving the possibility to apply basic mathematical
    operations on multiple results
    """

    def __init__(
        self,
        mesh_file="",
        domain=None,
        elem=("Lagrange", 1),
        V=None,
        vout=None,
        comm=COMM_WORLD,
    ):
        """
        initialisation of the FEMParameters:

        Parameters
        ----------
        mesh_file       :str
            mesh directory and file name: by default ""
        domain       : None or mesh
            mesh domain on which the result is defined, by default None
        elem            :tupple (str, int)
            if None, ("Lagrange", 1), else (element type, element order), by default None
        V       : None or dolfinx.fem.functionspace
            functionspace on which the result is defined, by default None
        vout       : None or dolfinx.fem.Function
            Function resulting from the FEMSimulation, by default None
        comm            :int
            The MPI communicator to use for mesh creation, by default MPI.COMM_WORLD
        """
        super().__init__()
        self.type = "simresult"
        self.mesh_file = mesh_file
        self.domain = domain
        self.V = V
        self.elem = elem
        self.vout = vout
        self.comm = comm
        
        #For eval() serialized calls acceleration
        self.is_evaluated = False
        self.tdim = None
        self.n_entities_local = None
        self.entities = None
        self.midpoint_tree = None
        self.tree = None


    def set_sim_result(
        self, mesh_file="", domain=None, V=None, elem=None, vout=None, comm=None
    ):
        if mesh_file != "":
            self.mesh_file = mesh_file
        if domain is not None:
            self.domain = domain
        if elem is not None:
            self.elem = elem
        if V is not None:
            self.V = V
        if vout is not None:
            self.vout = vout
        self.comm = comm

    def save(self, file, ftype="vtx", overwrite=True, t=0.0):
        fname = rmv_ext(file)
        if ftype.lower() == "vtx":  # and dfx_utd:
            fname += ".bp"
            with VTXWriter(self.domain.comm, fname, self.vout, "bp4") as f:
                f.write(t)
        elif ftype == "xdmf":
            fname += ".xdmf"
            with XDMFFile(self.comm, fname, "w") as file:
                if not overwrite:
                    file.parameters.update(
                        {"functions_share_mesh": True, "rewrite_function_mesh": False}
                    )
                else:
                    file.write_mesh(self.domain)
                file.write_function(self.vout)
        else:
            fname += ".sres"
            mdict = {
                "mesh_file": self.mesh_file,
                "element": self.elem,
                "vout": self.vector,
            }
            scipy.io.savemat(fname, mdict)
            return mdict

    def load(self, file):
        fname = rmv_ext(file) + ".sres"
        mdict = scipy.io.loadmat(fname)
        self.mesh_file = mdict["mesh_file"][0]
        self.elem = (mdict["element"][0].strip(), int(mdict["element"][1]))
        if self.domain is None:
            self.domain = domain_from_meshfile(self.mesh_file)
            self.V = functionspace(self.domain, self.elem)
        self.vout = Function(self.V)
        self.vout.x.array[:] = mdict["vout"]

    def save_sim_result(self, file, ftype="vtx", overwrite=True):
        rise_warning("save_sim_result is a deprecated method use save")
        self.save(file=file, ftype=ftype, overwrite=overwrite)

    def load_sim_result(self, data="sim_result.json"):
        rise_warning("load_sim_result is a deprecated method use load")
        # self.load(data=data)
        # FIXME: data does not exist yet in load method
        self.load(file=data)

    #############
    ## methods ##
    #############
    @property
    def vector(self):
        return self.vout.x.array

    def aline_V(self, res2):
        """
        Change the function space of the result to aline with the result of another FEMResults

        Parameters
        ----------
        res2 : FEMResults
            result to aline with
        """
        if is_sim_res(res2):
            if self.mesh_file == res2.mesh_file:
                _vout = Function(res2.V)
                _vout.interpolate(self.vout)
                self.vout=_vout
                self.V = res2.V
            else:
                rise_error(
                    "To aline mesh function reslults must have the same meshfile"
                )
        else:
            rise_error("Mesh function alinment must be done with FEMResults")

    def clone_res(self):
        return FEMResults(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )

    def eval(self, X, is_multi_proc=False):
        """
        Eval the result field at X position
        """
        X = np.array(X)
        N = len(X)
        cells = []
        points_on_proc = []
        if self.is_evaluated is False:
            self.is_evaluated = True
            self.tdim = self.domain.geometry.dim
            self.n_entities_local = (
            self.domain.topology.index_map(self.tdim).size_local
            + self.domain.topology.index_map(self.tdim).num_ghosts
            )
            self.entities = np.arange(self.n_entities_local, dtype=np.int32)
            self.midpoint_tree = create_midpoint_tree(self.domain, self.tdim, self.entities)
            self.tree = bb_tree(self.domain, self.tdim)
        # Find cells whose bounding-box collide with the the points

        cells_candidates = compute_collisions_points(self.tree, X)
        # Choose one of the cells that contains the point
        cells_colliding = compute_colliding_cells(self.domain, cells_candidates, X)
        for i in range(N):
            cell = cells_colliding.links(i)
            # point not in the mesh
            if len(cell) == 0:
                cell, x_closest = closest_point_in_mesh(
                    self.domain, X[i], self.tree, self.tdim, self.midpoint_tree
                )
                rise_warning(
                    X[i], " not found in mesh, value of ", x_closest, " reused"
                )
                # compute_colliding_cells(self.domain, cells_candidates, X)
                cells += [cell[0]]
            else:
                cells += [cell[0]]
        values = self.vout.eval(X, cells)
        if N > 1:
            values = values[:, 0]
        return values

    #####################
    ## special methods ##
    #####################
    def __abs__(self):
        res = self.clone_res()
        if self.vout is not None:
            res.vout = Function(self.V)
            res.vout.x.array[:] = abs(self.vout.x.array)
        return res

    def __neg__(self):
        res = self.clone_res()
        if self.vout is not None:
            res.vout = Function(self.V)
            res.vout.x.array[:] = -self.vout.x.array
        return res

    def __add__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = A + B
        return C

    def __sub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = A - B
        return C

    def __mul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = A * B
        return C

    def __radd__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = B + A
        return C

    def __rsub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = B - A
        return C

    def __rmul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            B = b.vout.x.array
        else:
            B = b
        A = self.vout.x.array
        C = self.clone_res()
        C.vout = Function(self.V)
        C.vout.x.array[:] = B * A
        return C

    def __eq__(self, b):
        if is_sim_res(b):
            if b.mesh_file == self.mesh_file:
                if np.allclose(self.vector, b.vector):
                    return True
        return False

    def __ne__(self, b):  # self != b
        return not self == b



"""
Everything about the configuration, gathers :in a singleton:
* Macine configuration,
* NRV parameters,
* ...
"""
from ._NRV_Singleton import NRV_singleton
from ._machine_config import MachineConfig
from ._parameters import nrv_parameters

class nrv_config(metaclass=NRV_singleton):
    """ A unique class to handle all the configuration
    
    """
    def __init__(self):
        self.machine_config = MachineConfig()
        self.framework_parameters = nrv_parameters()

    def display_machine_config(self):
        print(self.machine_config)


"""
NRV-:class:`.Problem` handling.
"""

import numpy as np
import faulthandler
import traceback

from ..backend._parameters import parameters
from ..backend._NRV_Class import NRV_class
from ..backend._log_interface import rise_error, pass_debug_info, set_log_level
from .optim_utils._OptimResults import optim_results
from ._CostFunctions import cost_function
from ._Optimizers import Optimizer

import sys

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()

class Problem(NRV_class):
    """
    Problem Class

    A class to describe problems that should be optimized with the NRV Framework.
    The problem should be described with a simulation and a cost, using the object
    cost_function, and various optimization algorithms can be used to find optimal solution.

    This class is abstract and is not supposed to be used directly by the end user. NRV can
    handle two types of problems:
    - problems where a geometric parameter can be optimized: please refer to ...
    - problems where the waveform can be optimized: please refer to ...
    """

    def __init__(
        self,
        cost_function: cost_function = None,
        optimizer: Optimizer = None,
        save_problem_results=False,
        problem_fname="optim.json",
    ):
        super().__init__()
        self._CostFunction = cost_function
        self._Optimizer = optimizer
        # For cases where optimisation is done on a swarm(groupe) of particle
        self.swarm_optimizer = False
        # self._SwarmCostFunction = None
        self.save_problem_results = save_problem_results
        self.problem_fname = problem_fname
        self.mp_type = None

    # Handling the cost_function attribute
    @property
    def costfunction(self):
        """
        Cost function of a Problem,
        the cost function should be a CosFunction object, it should return a scalar.
        NRV function should be prefered
        """
        return self._CostFunction

    @costfunction.setter
    def costfunction(self, cf: cost_function):
        # need to add a verification that the cost function is a scallar and so on
        self._CostFunction = cf

    
    @costfunction.deleter
    def costfunction(self):
        self._CostFunction = None

    def _SwarmCostFunction(self, swarm):
        s_l = len(swarm)
        costs = np.zeros((s_l))
        for i in range(s_l):
            particle = swarm[i][:]
            costs[i] = self._CostFunction(particle)
        return costs

    def compute_cost(self, X):
        return self._CostFunction(X)

    # Handling the Optimizer attribute
    @property
    def optimizer(self):
        """
        Optimizer of the problem,
        the Optimizer should be an Optimizer object. It has reference to optimization
        methods and constraints
        """
        return self._Optimizer

    @optimizer.setter
    def optimizer(self, optim: Optimizer):
        self._Optimizer = optim
        self.swarm_optimizer = self._Optimizer.swarm_optimizer

    @optimizer.deleter
    def optmizer(self):
        # self._Optimizer = None
        pass

    # Call method is where the magic happens
    def __call__(self, **kwargs) -> optim_results:
        """
        Perform the optimization: minimze the `cost_function` using `optmizer`

        Parameters
        ----------

        kwargs
            containing parameters of the optimizer to change

        Returns
        -------
        optim_results
            results of the optimization


        Raises
        ------
        KeyboardInterrupt
        """
        try:
            kwargs = self.__update_saving_parameters(**kwargs)
            if self.mp_type is None:
                self.set_multiprocess_type()
            if not self.swarm_optimizer:
                results = self._Optimizer(self._CostFunction, **kwargs)
            else:
                results = self._Optimizer(self._SwarmCostFunction, **kwargs)
                results["status"] = "Completed"
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except:
            results["status"] = "Error"
            rise_error(traceback.format_exc())


        set_log_level("INFO")
        if self.save_problem_results:
            results.save(save=True, fname=self.problem_fname)
        return results


    # Mcore handling
    def __check_MCore_CostFunction(self):
        """
        check if a cost funciton can be parallelized
        """
        return getattr(self._CostFunction, "_MCore_CostFunction", False)

    def set_multiprocess_type(self, costfunction_mp=True, n_core=None):
        """
        Set if multiprocessing should be applied to the optimizaiton or the CostFunction simulation

        Warning
        -------
        For now, only costfunction can be parallelized. This will be improve in the future
        """
        if n_core is None:
            n_core = parameters.get_optim_ncore()
        # parallelizable optimizer
        if "n_processes" in self._Optimizer.__dict__:
            if self.__check_MCore_CostFunction() and costfunction_mp:
                #* To add number of n_core_fascicle = n_core
                self._Optimizer.n_processes = None
                self.mp_type = "costfunction"
            else:
                #* To add number of n_core_fascicle = 1
                #!! Bug cannot compute local method generated from cost_function_swarm_from_particle
                self._Optimizer.n_processes = parameters.get_optim_ncore()
                #!!self._Optimizer.n_processes = None
                self.mp_type = "optimizer"
        else:
            #* To add number of n_core_fascicle = n_core
            self.mp_type = "costfunction"


    # additional methods
    def __update_saving_parameters(self, **kwargs):
        """
        internal use only: update the results saving parameters
        and remove the corresponding keys from kwargs
        """
        if "save_problem_results" in kwargs:
            self.save_problem_results = kwargs.pop("save_problem_results")
        if "problem_fname" in kwargs:
            self.problem_fname = kwargs.pop("problem_fname")
        return kwargs

    def context_and_cost(self, context_func, cost_func, residual):
        self.cost_function = cost_function(context_func, cost_func, residual)

    def autoset_optimizer(self):
        pass



"""
NRV-fenics_materials class handling.
"""

import faulthandler
import os

import numpy as np

from ....backend._file_handler import json_dump, rmv_ext
from ....backend._log_interface import rise_warning
from ....utils._nrv_function import nrv_interp, nrv_function
from ..._materials import (
    is_mat,
    load_material,
    material,
    compute_effective_conductivity,
)

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()


####################
## material class ##
####################
class f_material(material):
    """
    a class for conductive material wh

    parameters
    ----------
    mat     :material
        generate the fenics material from mat attribute
    """

    def __init__(self):
        """
        initialisation of the fenics_material
        """
        super().__init__()
        self._sigma_func = None

    ## Save and Load mehtods
    def save(self, save=False, fname="Fenics_model.json", blacklist=[], **kwargs):
        """
        Return material as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "material.json"

        Returns
        -------
        mat_dic : dict
            dictionary containing all information
        """
        bl = [i for i in blacklist]
        bl += ["sigma_func"]
        return super().save(save=save, fname=fname, blacklist=bl, **kwargs)

    @property
    def is_func(self):
        return not self._sigma_func is None

    @property
    def sigma(self):
        if self.is_func:
            return self.sigma_func
        else:
            return super().sigma

    @property
    def sigma_func(self):
        return compute_effective_conductivity(
            sigma=self._sigma_func, epsilon=self.epsilon, freq=self.freq
        )

    def set_conductivity_function(self, sigma_func: nrv_function) -> None:
        """
        set the conductivity space function for an anisotropic material

        Parameters
        ----------
        sigma_fuction    : func(X : array([3, N]))-> array([N])
            conductivity function in 3D space
        """
        self.isotrop_cond = False
        self._sigma_func = sigma_func

    def is_function_defined(self) -> bool:
        """
        check that the material conductivity is define as a function

        Returns
        -------
        bool    :
            True if the per
        """
        return self.is_func


###############
## Functions ##
###############


def is_f_mat(mat: object) -> bool:
    """
    check if an object is a fenics_material, return True if yes, else False

    Parameters
    ----------
    mat : object
        object to test

    Returns
    -------
    bool
        True it the type is a material object
    """
    return isinstance(mat, f_material)


def mat_from_interp(
    X, Y, kind="linear", dx=0.01, interpolator=None, dxdy=None, scale=None, columns=0
) -> f_material:
    """
    Return a fenics material with a conductivity space function define as the
    :class:`~nrv.utils.nrv_function.nrv_interp` of X and Y

    Parameters
    ----------
    f_material  : str
        either material name if the material is in the NRV2 Librairy or path to the corresponding
        .mat material file
    kwargs      : dict
        interpolation k arguments (see :class:`~nrv.utils.nrv_function.nrv_interp`)

    Returns
    -------
    mat_obj     : fenics_material
        Fenics material with conductivity function defined from the interpolation
    """
    sigma_func = nrv_interp(
        X,
        Y,
        kind=kind,
        dx=dx,
        interpolator=interpolator,
        dxdy=dxdy,
        scale=scale,
        columns=columns,
    )

    # creat material instance
    mat_obj = f_material()
    mat_obj.set_conductivity_function(sigma_func)
    return mat_obj


def mat_from_csv(fname: str, **kwargs) -> f_material:
    """
    Return a fenics material with a conductivity space function define as the
    :class:`~nrv.utils.nrv_function.nrv_interp` from the value of a .csv file

    Parameters
    ----------
    f_material  : str
        name of the .csv file

    Returns
    -------
    mat_obj     : fenics_material
        Fenics material with conductivity function defined from the interpolation from the
        value of f_material file
    """
    fname = rmv_ext(fname) + ".csv"

    data = np.loadtxt(fname, delimiter=",")
    X = data[0]
    Y = data[1]

    return mat_from_interp(X, Y, **kwargs)


def load_f_material(X: any = None, **kwargs) -> f_material:
    """
    Return fenics material from an object X

    Parameters
    ----------
    X   : objects
        if material or fenics material returning the corresponding fenics material
        if float returns an isotropic fenics material with X as conductivity value
        if str finishing with .csv use mat_from_csv with kwargs
        else return None
    """
    mat = None
    if X is None:
        mat = f_material()
    elif is_f_mat(X):
        mat = X
    elif is_mat(X):
        mat = f_material()
        mat.load(X.save(save=False))
    elif isinstance(X, str):
        if ".csv" in X:
            mat = mat_from_csv(X, **kwargs)
        else:
            mat = f_material()
            mat.load(load_material(X).save(save=False))
    elif isinstance(X, (complex, float, int)):
        mat = f_material()
        mat.set_isotropic_conductivity(X)
    elif np.iterable(X):
        if len(X) == 2:
            mat = mat_from_interp(X[0, :], X[1, :], **kwargs)
        elif len(X) == 3:
            mat = f_material()
            mat.set_anisotropic_conductivity(X[0], X[1], X[2])
    elif hasattr(X, "mat"):
        mat = load_f_material(X.mat)
    if mat is None:
        rise_warning(
            TypeError(),
            f"{type(X)} not convertible in f_material\n",
            "empty material generated",
        )
        mat = f_material()
    return mat



from ..backend._NRV_Class import NRV_class
from ..backend._log_interface import rise_error


class eit_forward(NRV_class):
    """
    Class allowing to simulate Electircal Impedance Tomography in a nerve
    """

    def __init__(self):
        super().__init__()
        self._nerve = None
        self._protocol = None

    @property
    def nerve(self):
        return self._nerve

    @nerve.setter
    def nerve(self, n):
        self._nerve = n

    @nerve.deleter
    def nerve(self, n):
        self._nerve = None

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, p):
        self._protocol = p

    @protocol.deleter
    def protocol(self, p):
        self._protocol = None

    def simulate():
        rise_error(NotImplementedError)



"""
NRV-:class:`.myelinated_results` handling.
"""

import numpy as np
import matplotlib.pyplot as plt
from ._axons_results import axon_results
from ...backend._log_interface import rise_warning
from ...fmod._materials import is_mat, load_material
from ...utils._units import to_nrv_unit, convert


class myelinated_results(axon_results):
    """ """

    def __init__(self, context=None):
        super().__init__(context)

    def generate_axon(self):
        if "myelinated" not in globals():
            from .._myelinated import myelinated
        return myelinated(**self)

    def get_index_myelinated_sequence(self, n):
        """
        Returns the simulated myelination sequence of the axon corresponding to a calculation
        point index.

        Parameters
        ----------
        n     : int
            intex to check.

        Returns
        -------
        str
            corresponding sequence.
        """
        if self["rec"] == "nodes":
            return "node"
        else:
            if n > len(self["x_rec"]):
                rise_warning("index not in axon")
            # +1 required because nbr of computation point = nbr seg/sec + 1
            # see if it's a bug
            Nseg_per_sec = self["Nseg_per_sec"] + 1
            N_sec_type = 11
            seq_types = self["axon_path_type"]
            if n == 0:
                return seq_types[0]
            else:
                return seq_types[((n - 1) // Nseg_per_sec) % N_sec_type]

    def find_central_node_coordinate(self, node=True):
        """
        Returns the index of the closer node from the center

        Returns
        -------
        float
            x-position of the closer node from the center
        """
        return self["x_rec"][self.find_central_index(node=node)]

    def find_central_index(self, node:bool=True)->int:
        """
        Returns the central indec

        Parameters
        ----------
        node : bool
            If true, the index of the closer node from the center is return, by default True

        Returns
        -------
        int
            index of `x_rec` of the closer node from the center
        """
        n_center = super().find_central_index()
        if node and self["rec"] == "all":
            for i in range(n_center):
                if self.get_index_myelinated_sequence(n_center + i) == "node":
                    return n_center + i
                elif self.get_index_myelinated_sequence(n_center - i) == "node":
                    return n_center - i
            rise_warning("No node found in the axon")
        return n_center

    def get_myelin_properties(self, endo_mat=None):
        """
        compute the cutoff frequency of the axon's membrane and add it to the simulation results dictionnary
        NB: The frequency is computed in [kHz]

        Returns
        -------
        g_mye              : np.ndarray
            value of the cutoff conductivity of the axon's membrane
        c_mye              : np.ndarray
            value of the cutoff capacitance of the axon's membrane
        f_mye              : np.ndarray
            value of the cutoff frequency of the axon's membrane
        """
        if self["rec"] == "nodes":
            rise_warning("No myelin in the axon simulated, None returned")
            return None

        ax = self.generate_axon()
        self["g_mye"] = ax.get_myelin_conductance()
        if endo_mat is not None:
            if not is_mat(endo_mat):
                endo_mat = load_material(endo_mat)
            I = np.isclose(self["g_mye"], 1e10)
            self["g_mye"][I] *= 0.0
            self["g_mye"][I] += convert(endo_mat.sigma, "S/m**2", "S/cm**2")
        self["c_mye"] = ax.get_myelin_capacitance()
        self["f_mye"] = self["g_mye"] / (2 * np.pi * self["c_mye"])

        # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
        # * [MHz] to convert to [kHz]
        self["f_mye"] = to_nrv_unit(self["f_mye"], "MHz")
        return self["g_mye"], self["c_mye"], self["f_mye"]

    def plot_x_t(
        self, axes: plt.axes, key: str = "V_mem", color: str = "k", switch_axes=False, norm=None, **kwgs
    ) -> None:
        x_pos = self.x[self.node_index]
        x_index = self.node_index
        if not "ALL" in self.rec.upper():
            x_index = np.arange(len(x_pos))
        super().plot_x_t(axes=axes, x_pos=x_pos,x_index=x_index, key=key, color=color, switch_axes=switch_axes, norm=norm, **kwgs)

    def plot_x_t_all_seq(
        self, axes: plt.axes, key: str = "V_mem", color: str = "k", myelin_color:str="gray", switch_axes=False, norm=None, **kwgs
    ) -> None:
        if self.rec=="all":
            node_index = self.node_index
            other_index = np.concatenate([i_x+(self.Nseg_per_sec+1)*np.arange(len(self.this_ax_sequence)) for i_x in node_index])
            other_index = other_index[other_index<len(self.x_rec)]
            super().plot_x_t(axes=axes, x_pos=self.x_rec[other_index],x_index=other_index, key=key, color=myelin_color, switch_axes=switch_axes, norm=norm, **kwgs)

        self.plot_x_t(axes=axes, key=key, color=color, switch_axes=switch_axes, norm=norm)


    def colormap_plot(self, axes: plt.axes, key: str = "V_mem", **kwgs) -> plt.colorbar:
        if self.rec == "all":
            return super().colormap_plot(axes, key, **kwgs)
        else: 
            rise_warning("colormap_plot be myelined.rec is not 'all'")
            return None
    

    def plot_Nav_states(ax: plt.axes) -> None:
        """
        Plot the state machine for kinetic (Markov) Nav 1.1 to 1.9 values

        Parameters
        ----------
        ax : matplotlib axis object
            axes of the figure to work on
        """

        values = []
        rise_warning("plot_Nav_states is not fully implemented")
        states = [r"$I_1$", r"$I_2$", r"$C_1$", r"$C_2$", r"$O_1$", r"$O_2$"]

        X = [-1, -3, 0, 1, 0, 3]
        Y = [0, 0, 1, 0, -1, 0]
        c = ["r", "r", "b", "b", "g", "g"]

        ax.set_xlim(-3.4, 3.4)
        ax.set_ylim(-1.5, 1.5)
        for i in range(len(states)):
            ax.scatter(X[i], Y[i], s=300 + values[i] * 1450, c=c[i], alpha=0.4)
            ax.text(X[i], Y[i], states[i], ha="center", va="center")
        # paths
        ax.arrow(
            -2.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
        )
        ax.text(-2, 0.2, "$I_2I_1$", ha="center", va="center", alpha=0.4)
        ax.arrow(
            -1.5,
            -0.03,
            -1,
            0,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(-2, -0.2, "$I_1I_2$", ha="center", va="center", alpha=0.4)

        ax.arrow(
            -0.83,
            0.25,
            0.5,
            0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(-0.9, 0.6, "$I_1C_1$", ha="center", va="center", alpha=0.4)
        ax.arrow(
            -0.22,
            0.75,
            -0.5,
            -0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(-0.45, 0.25, r"$C_1I_1$", ha="center", va="center", alpha=0.4)

        ax.arrow(
            0.72,
            0.25,
            -0.5,
            0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(0.9, 0.6, r"$C_1C_2$", ha="center", va="center", alpha=0.4)
        ax.arrow(
            0.33,
            0.75,
            0.5,
            -0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(0.45, 0.25, r"$C_2C_1$", ha="center", va="center", alpha=0.4)

        ax.arrow(
            0.83,
            -0.25,
            -0.5,
            -0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(0.9, -0.6, r"$C_2O_1$", ha="center", va="center", alpha=0.4)
        ax.arrow(
            0.22,
            -0.75,
            0.5,
            0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(0.45, -0.25, r"$O_1C_2$", ha="center", va="center", alpha=0.4)

        ax.arrow(
            -0.33,
            -0.75,
            -0.5,
            0.5,
            linewidth=1,
            alpha=0.5,
            head_width=0.02,
            head_length=0.02,
        )
        ax.text(-0.9, -0.6, r"$O_1I_1$", ha="center", va="center", alpha=0.4)

        ax.arrow(
            1.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
        )
        ax.text(2, 0.2, r"$C_20_2$", ha="center", va="center", alpha=0.4)
        ax.arrow(
            2.5, -0.03, -1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
        )
        ax.text(2, -0.2, r"$O_2C_2$", ha="center", va="center", alpha=0.4)
        # make axes to disappear
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")



from ..backend._NRV_Class import NRV_class
from copy import deepcopy
import numpy as np


class protocol(NRV_class):
    """
    Caution not fully implemented..
    First version of a eit protocol class
    """

    def __init__(self, p=None):
        super().__init__()
        self._injection = []
        self._recording = []
        self.pos = 0
        if isinstance(p, protocol):
            self._injection = deepcopy(p._injection)
            self._recording = deepcopy(p._recording)

    @property
    def rec_mat(self):
        m_mat = [[]]
        for i in range(len(self) - 1):
            m_mat[-1] += [self._recording[i]]
            if self._injection[i] != self._injection[i + 1]:
                m_mat += [[]]
        m_mat[-1] += [list(self._recording[-1])]
        return np.array(m_mat)

    @property
    def inj_mat(self):
        e_mat = [self._injection[0]]
        for i in range(len(self) - 1):
            if self._injection[i] != self._injection[i + 1]:
                e_mat += [self._injection[i + 1]]
        return np.array(e_mat)

    def change_electrode_id(self, oldid: int, newid: int):
        for i, e in enumerate(self._injection):
            e1, e2 = e
            if oldid == e[0]:
                e1 = newid
            if oldid == e[1]:
                e2 = newid
            self._injection[i] = e1, e2
        for i, e in enumerate(self._recording):
            e1, e2 = e
            if oldid == e[0]:
                e1 = newid
            if oldid == e[1]:
                e2 = newid
            self._recording[i] = e1, e2

    def add_injection(self, inj=None):
        if inj is not None:
            self._injection += [inj]
        else:
            self._injection += [self._injection[-1]]

    def add_recording(self, rec=None):
        if rec is not None:
            self._recording += [rec]
        else:
            self._recording += [self._recording[-1]]

    def add_patern(self, inj=None, rec=None):
        if len(self) == 0 and (inj is None or rec is None):
            raise IndexError("empty ") from None
        else:
            self.add_injection(inj)
            self.add_recording(rec)

    def clear(self):
        self._injection = []
        self._recording = []

    def push(self, inj=None, rec=None):
        self.add_patern(inj=inj, rec=rec)

    def pop(self):
        try:
            return self._injection.pop(), self._recording.pop()
        except IndexError:
            raise IndexError("pop from an empty stack") from None

    def __getitem__(self, index):
        return (self._injection[index], self._recording[index])

    def __len__(self):
        return len(self._injection)

    def add_n(self, n: int):
        if n + np.min(self) < 0:
            print("Warning: protocol cannot have negative id of electrode")
            print(f"Warning: protocol cannot have negative id of electrode")
            n = np.min(self)

        self._injection = [(n + i, n + j) for i, j in self._injection]
        self._recording = [(n + i, n + j) for i, j in self._recording]

    def mul_n(self, n: int):
        self._injection = [(n * i, n * j) for i, j in self._injection]
        self._recording = [(n * i, n * j) for i, j in self._recording]

    def __sub__(self, n: int):
        """ """
        p = protocol(self)
        p.add_n(-n)
        return p

    def __rsub__(self, n: int):
        return self.__add__(-n)

    def __add__(self, n: int):
        """ """
        p = protocol(self)
        p.add_n(n)
        return p

    def __radd__(self, n: int):
        return self.__add__(n)

    def __mul__(self, n: int):
        """ """
        p = protocol(self)
        p.mul_n(n)
        return p

    def __rmul__(self, n: int):
        return self.__add__(n)


class pyeit_protocol(protocol):
    def __init__(self, n_elec=8, inj_offset=1, start_elec=0):
        super().__init__()
        self.n_elec = n_elec
        self.inj_offset = inj_offset
        self.start_elec = start_elec
        self.__generate_protocol()

    def __generate_protocol(self):
        self.clear()
        for i_inj in range(self.n_elec):
            inj_pat = (
                (self.start_elec + i_inj) % self.n_elec,
                (self.start_elec + i_inj + self.inj_offset) % self.n_elec,
            )
            inj_pat = inj_pat[0], inj_pat[1]
            for i_rec in range(self.n_elec):
                e = (self.start_elec + i_rec + 1) % self.n_elec
                e_ = (self.start_elec + i_rec) % self.n_elec
                if not e in inj_pat and not e_ in inj_pat:
                    rec_pat = e, e_
                    if not (rec_pat[0] in inj_pat or rec_pat[1] in inj_pat):
                        self.add_patern(inj_pat, rec_pat)
        self = self.add_n(1)

    def set_parameters(self, **kwds):
        for key in kwds:
            if key in self.__dict__:
                self.__dict__[key] = kwds[key]
        self.__generate_protocol()



import numpy as np

from ...utils._stimulus import stimulus, set_common_time_series
from ...fmod._extracellular import extracellular_context

from ...backend._NRV_Class import NRV_class, load_any, abstractmethod
from ...backend._NRV_Simulable import sim_results
from ...nmod.results._axons_results import axon_results
from ...nmod.results._fascicles_results import fascicle_results
from ...nmod.results._nerve_results import nerve_results
from ...utils._nrv_function import cost_evaluation
from ...ui._axon_postprocessing import *


class raster_count_CE(cost_evaluation):
    """
    Create a callable object which returne the number of spike from the result
    of a simulation
    """

    def __init__(self):
        super().__init__()

    def call_method(self, results: sim_results, **kwargs) -> float:
        """
        Returns the spike number from a simulation result
        """
        if "V_mem_raster_position" not in results:
            results.rasterize("V_mem")
        pos = results["V_mem_raster_position"]
        M = len(results["x_rec"]) - 1  # pos starts at 0
        i_first_pos = np.where(pos == 0)
        i_last_pos = np.where(pos == M)
        cost = (len(i_first_pos[0]) + len(i_last_pos[0])) / 2
        return cost


class recrutement_count_CE(cost_evaluation):
    r"""
    Callable object which returns the number of triggered fibre in the results

    Parameters
    ----------
    reverse     : bool
        if True, the final cost is the difference between the number total of fibre and the number of activate fibre

    Note
    ----
    if reverse is false:

    .. math::

        cost = N_{recruited}

    else:

    .. math::

        cost = N_{total} - N_{recruited}
    """

    def __init__(self, reverse=False):
        super().__init__()
        self.reverse = reverse

    def count_axon_activation(self, results: sim_results):
        cpt = results.is_recruited()
        if self.reverse:
            cpt = not cpt
        return int(cpt)

    def count_fascicle_activation(self, results: sim_results):
        cpt = 0
        for i in range(len(results["axons_diameter"])):
            if self.reverse:
                cpt += 1 - results["axon" + str(i)]["spike"]
            else:
                cpt += results["axon" + str(i)]["spike"]
        return cpt

    def call_method(self, results: sim_results, **kwargs) -> float:
        """
        Returns the spike number from a simulation result

        Parameters
        ----------
        results     : dict
            output of an axon simulation using Markov model for at least a node

        Returns
        -------
        cost        :int
            number of spike in the v_mem part
        """
        cost = 0
        if isinstance(results, axon_results):
            cost = self.count_axon_activation(results)
        else:
            cost = results.get_recruited_axons(ax_type="all", normalize=False)
            if self.reverse:
                cost = results.n_ax - cost
        return cost


class charge_quantity_CE(cost_evaluation):
    r"""
    Create a callable object which return a value proportionnal to the charge quantity injected by stimulus.

    .. math::

        cost = \sum_{e}\sum_{t_k}{i_{e,stim}(t_k)}

    with :math:`t_k` is the discrete time step of the simulation
    """

    def __init__(self, id_elec=None, dt_res=0.0001):
        super().__init__()
        self.id_elec = id_elec
        self.dt_res = dt_res

    def compute_stimulus_cost(self, stim: stimulus):
        stim_ = stimulus()
        t_min, t_max = stim.t[0], stim.t[-1]
        N_pts = int((t_max - t_min) // self.dt_res)
        stim_.t = np.linspace(t_min, t_max, N_pts)
        set_common_time_series(stim, stim_)
        return abs(stim).integrate()

    def call_method(self, results: sim_results, **kwargs) -> float:
        extra_stim = load_any(results["extra_stim"])
        N_elec = len(extra_stim.stimuli)
        cost = 0
        if self.id_elec is None:
            self.id_elec = [k for k in range(N_elec)]
        elif isinstance(self.id_elec, int):
            self.id_elec = [self.id_elec]
        for i in self.id_elec:
            cost += self.compute_stimulus_cost(extra_stim.stimuli[i])
        return cost


class stim_energy_CE(cost_evaluation):
    r"""
    Create a callable object which return a value proportionnal to the stimulus energy, assuming the electrode impedance is a constant.

    .. math::

        cost = \sum_{e}\sum_{t_k}{i_{e,stim}^2(t_k)}

    with :math:`t_k` is the discrete time step of the simulation

    Parameters
    ----------
    id_elec : None | int | list[int]
        id or list id of the electrode of the to from which the energy should be computed. If None,
    dt_res  : float
        resolotion time step use to compute the cost value
    """

    def __init__(self, id_elec: None | int | list[int] = None, dt_res: float = 0.0001):
        super().__init__()
        self.id_elec = id_elec
        self.dt_res = dt_res

    def compute_stimulus_cost(self, stim: stimulus):
        stim_ = stimulus()
        t_min, t_max = stim.t[0], stim.t[-1]
        N_pts = int((t_max - t_min) // self.dt_res)
        stim_.t = np.linspace(t_min, t_max, N_pts)
        set_common_time_series(stim, stim_)
        stim.s = stim.s * stim.s
        return abs(stim).integrate()

    def call_method(self, results: sim_results, **kwargs) -> float:
        extra_stim = load_any(results["extra_stim"])
        N_elec = len(extra_stim.stimuli)
        cost = 0
        if self.id_elec is None:
            self.id_elec = [k for k in range(N_elec)]
        elif isinstance(self.id_elec, int):
            self.id_elec = [self.id_elec]
        for i in self.id_elec:
            cost += self.compute_stimulus_cost(extra_stim.stimuli[i])
        return cost



from collections.abc import Callable
import inspect

from ._log_interface import rise_warning


def set_attributes(my_object, attributes_dict):
    for key, value in attributes_dict.items():
        if key in my_object.__dict__:
            setattr(my_object, key, value)
        else:
            rise_warning(
                "trying to set a non existing attribute "
                + str(key)
                + " in "
                + str(type(my_object))
            )
    return 0


def check_function_kwargs(func: Callable, kwargs: dict) -> dict:
    """
    check that the keys of a dictionnary are arguments of a function and return an updated dictionnary with only valide keys

    Parameters
    ----------
    func : function
        function to check
    kwargs : dict
        dictionnary of arguments to check

    Returns
    -------
    dict
        updated kwargs dictionnary
    """
    func_kwargs_set = set(func.__code__.co_varnames[: func.__code__.co_argcount])
    kwargs_set = set(kwargs.keys())
    not_valid_kwargs_set = kwargs_set - func_kwargs_set
    for k in not_valid_kwargs_set:
        kwargs.pop(k)
    return kwargs


def function_to_str(func: Callable) -> str:
    lines = inspect.getsource(func)
    return lines


def str_to_function(lines: str) -> Callable:
    lines
    return lines



r"""
The ``units`` module handles conversion to and from the default nrv units.
Several functions are included to simplify the use of these units.

The folowing table details the default units used in NRV and the lists of units available for conversion:
    .. list-table::
        :align: left
        :header-rows: 1

        *   - Category
            - Default
            - Existing
        *   - Voltage
            - mV
            - uV, V, kV, MV
        *   - Time
            - ms
            - us, s, minute, hour, day
        *   - Length
            - um
            - nm, mm, cm, dm, m
        *   - Currents
            - uA
            - nA, mA, A
        *   - Conductance
            - S
            - kS, mS, uS
        *   - Frequency
            - kHz
            - Hz, MHz, GHz
        *   - Capacitance
            - uF
            - F, mF, pF, nF

"""

import numpy as np
from copy import deepcopy
from math import isnan

###################
## defaulf units ##
###################
mV = 1  # default unit for Voltage is mV, as neuron
ms = 1  # default unit for time is ms, as neuron
um = 1  # default unit for length is um, as neuron
uA = 1  # default unit for current is uA, UNLESS NEURON who is by default in nA
S = 1  # default unit for conductance is S
kHz = 1  # default unit for frequency is kHz
uF = 1  # default unit for capacitance is uF

######################
## voltage prefixes ##
######################
uV = 1e-3 * mV
V = 1e3 * mV
kV = 1e3 * V
MV = 1e3 * kV

#############################
## time prefixes and units ##
#############################
us = 1e-3 * ms
s = sec = 1e3 * ms
minute = 60 * sec
hour = 60 * minute
day = 24 * hour

########################
## frequency prefixes ##
########################

Hz = 1e-3 * kHz
MHz = 1e3 * kHz
GHz = 1e6 * kHz

#####################
## length prefixes ##
#####################
nm = 1e-3 * um
mm = 1e3 * um
cm = 1e1 * mm
dm = 1e1 * cm
m = 1e1 * dm

######################
## current prefixes ##
######################
nA = 1e-3 * uA
mA = 1e3 * uA
A = 1e3 * mA

###########################
## conductance prefixes ##
###########################
kS = 1e3 * S
mS = 1e-3 * S
uS = 1e-6 * S

###########################
## capacitance prefixes ##
###########################
F = 1e6 * uF
mF = 1e3 * uF
pF = 1e-3 * uF
nF = 1e-6 * uF

#############################
####  Usefull functions ####
#############################


def print_default_nrv_unit():
    """
    Print all the default units in nrv.
    """
    print(
        "Voltage: mV",
        "\nTime: ms",
        "\nLength: um",
        "\nCurrents: uA",
        "\nConductance: S",
        "\nFrequency: kHz",
        "\nCapacitance: uF",
    )


def from_nrv_unit(value, unit):
    """
    Convert a quantity ``value`` from the default nrv units to ``unit``.

    Parameters
    ----------
    value:      int, float, list or np.ndarray
        value or values wich should be converted.
    unit:     int or str
        unit from which value should be converted, see example.

    Returns
    -------
    int, float, list or np.ndarray
        rounded value or values, with the same type and shape than ``value``.

    Example
    -------
    Here are three ways of converting 4530um (default unit) into 4.53mm:

        >>> import nrv
        >>> val_default = 4530 # um
        >>> val_default / nrv.mm
        4.53
        >>> nrv.from_nrv_unit(val_default, nrv.mm)
        4.53
        >>> nrv.from_nrv_unit(val_default, "mm")
        4.53
    """
    if isinstance(unit, str):
        unit = eval(unit)
    if np.iterable(value) and not isinstance(value, np.ndarray):
        cp_value = deepcopy(value)
        for i, num in enumerate(value):
            cp_value[i] = from_nrv_unit(num, unit)
        return cp_value
    else:
        return value / unit


def to_nrv_unit(value, unit):
    """
    Convert a quantity ``value`` from unit to the default nrv ``unit``.

    Parameters
    ----------
    value:      int, float, list or np.ndarray
        value or values wich should be converted.
    unit:     int or str
        unit to which value should be converted, see examples.

    Returns
    -------
    int, float, list or np.ndarray
        rounded value or values, with the same type and shape than ``value``.

    Example
    -------
    Here are three ways of converting 0.12 MHz into the default unit of the corresponding nrv (kHz):

        >>> import nrv
        >>> val_MHz = 0.12 # MHz
        >>> val_MHz * nrv.MHz
        120.0
        >>> nrv.to_nrv_unit(val_MHz, nrv.MHz)
        120.0
        >>> nrv.to_nrv_unit(val_MHz, "MHz")
        120.0

    Here is how to convert a ``list`` or ``np.array`` of values in seconds
    to milliseconds (nrv's default unit):

        >>> import numpy as np
        >>> vals_s = [[1, 2, 3], [4, 5, 6]]  # s
        >>> nrv.to_nrv_unit(vals_s, nrv.s)  # ms
        [[1000.0, 2000.0, 3000.0], [4000.0, 5000.0, 6000.0]]
        >>> nrv.to_nrv_unit(np.array(vals_s), nrv.s)  # ms
        array([[1000., 2000., 3000.],
            [4000., 5000., 6000.]])

    """
    if isinstance(unit, str):
        unit = eval(unit)
    if np.iterable(value) and not isinstance(value, np.ndarray):
        cp_value = deepcopy(value)
        for i, num in enumerate(value):
            cp_value[i] = to_nrv_unit(num, unit)
        return cp_value
    else:
        return value * unit


def convert(value, unitin, unitout):
    """
    Convert a quantity ``value`` from ``unitin`` to ``unitout``.

    Parameters
    ----------
    value:      int, float, list or np.ndarray
        value or values wich should be converted.
    unit:     int or str
        unit to which value should be converted, see examples.

    Returns
    -------
    int, float, list or np.ndarray
        rounded value or values, with the same type and shape than ``value``.)

    Example
    -------
    Here are two ways of converting 0.2 S/m^{2} into S/cm^{2}:

        >>> import nrv
        >>> val_S_m = 0.2 # S/m**2
        >>> nrv.convert(val_S_m, nrv.S/nrv.m**2, nrv.S/nrv.cm**2)
        2e-05
        >>> nrv.convert(val_S_m, "S/m**2", "S/cm**2")
        2e-05
    """
    if np.iterable(value) and not isinstance(value, np.ndarray):
        cp_value = deepcopy(value)
        for i, num in enumerate(value):
            cp_value[i] = convert(num, unitin, unitout)
        return cp_value
    else:
        return from_nrv_unit(to_nrv_unit(value, unitin), unitout)


def sci_round(value, digits=3):
    """
    Rounds one or several values to ``digits`` significant digits.

    Parameters
    ----------
    value:      int, float, list or np.ndarray
        value or values wich should be rounded.
    digits:     int
        number of significant digits to keep.

    Returns
    -------
    int, float, list or np.ndarray
        rounded value or values, with the same type and shape than ``value``.
    """
    if np.iterable(value):
        cp_value = deepcopy(value)
        for i, num in enumerate(value):
            cp_value[i] = sci_round(num, digits)
        return cp_value
    else:
        if not isnan(value):
            power = "{:e}".format(value).split("e")[1]
            return round(value, -(int(power) - digits + 1))
        else:
            return value
